/*	
	Copyright(c) 2010 Christian Riess <christian.riess@cs.fau.de>.

	This file may be licensed under the terms of of the GNU General Public
	License, version 3, as published by the Free Software Foundation. You can
	find it here: http://www.gnu.org/licenses/gpl.html
*/

#ifndef PIXEL_H
#define PIXEL_H

#include "data_traits.h"

#include <iostream>
#include <cassert>
#include <sstream>
#include <cstring>

template <typename ChannelT = unsigned char, int dim = 3>
class Pixel {
public:
	typedef DataTraits<ChannelT> traits_type;
	typedef typename traits_type::data_type dataT;
	typedef typename traits_type::aggregate_type aggregateT;

	Pixel() : active(true) {
		for (int i = 0; i < dim; i++) colors[i] = 0;
	}

	Pixel(const ChannelT &new_colors) : active(true) {
		for (int i = 0; i < dim; i++) {
			colors[i] = static_cast<ChannelT>(new_colors[i]);
		}
	}

	//! copy-constructor
	Pixel(Pixel<ChannelT, dim> &p) {
		memcpy(colors, p.getColors(), dim * sizeof(ChannelT));
//		for (unsigned int i = 0; i < dim; i++) colors[i] = static_cast<ChannelT>(p[i]);
		active = p.isActive();
	}

	//! copy-constructor
	Pixel(const Pixel<ChannelT, dim> &p) {
		const ChannelT *entries = p.getColors();
		for (unsigned int i = 0; i < dim; i++) colors[i] = entries[i];
		active = p.isActive();
	}

	virtual ~Pixel() { }

	/* returns the dimension of the pixels */
	virtual int getDim() const { return dim; }

	/** returns the pixel colors as one array */
	virtual ChannelT *getColors() { return colors; }
	virtual const ChannelT *getColors() const { return colors; }

	//! accessor for a pixel
	const ChannelT &operator[](unsigned int pos) const {
		assert(pos < dim);
		return colors[pos];
	}

	//! accessor for a pixel
	ChannelT &operator[](unsigned int pos) {
		assert(pos < dim);
		return colors[pos];
	}

	unsigned int getChannelDepth() const { return sizeof(ChannelT); }

	/** assigns the entries of the argument pixel to this pixel.
	    if the argument has lower dimension, this pixel's remaining
	    dimensions are set to zero */
	Pixel<ChannelT, dim> &operator=(Pixel<ChannelT, dim> &p) {
		// FIXME: is it possible to ask for a pixel of arbitrary type ChannelT
		// as an argument of this function?
		int dim2 = p.getDim();
		assert(dim >= dim2);
		for (int i = 0; i < dim2; i++) colors[i] = static_cast<ChannelT>(p[i]);
		for (int i = dim2; i < dim; i++) colors[i] = 0;
		active = p.isActive();
		return *this;
	}

	Pixel<ChannelT, dim> &operator=(const Pixel<ChannelT, dim> &p) {
		// FIXME: is it possible to ask for a pixel of arbitrary type ChannelT
		// as an argument of this function?
		int dim2 = p.getDim();
		assert(dim >= dim2);
		for (int i = 0; i < dim2; i++) colors[i] = static_cast<ChannelT>(p[i]);
		for (int i = dim2; i < dim; i++) colors[i] = 0;
		active = p.isActive();
		return *this;
	}

	void setZero() {
		for (int i = 0; i < dim; i++) { colors[i] = 0; }
	}

	void setZeroAndInactive() {
		for (int i = 0; i < dim; i++) { colors[i] = 0; }
		active = false;
	}

	template <typename D>
	void set(int pos, D val) {
		assert(pos >= 0); assert(pos < dim);
		colors[pos] = static_cast<ChannelT>(val);
	}

	std::string toString();


	virtual bool isActive() const { return active; };
	virtual bool isInactive() { return active == false; }

	virtual void setActive() { active = true; }
	virtual void setInactive() { active = false; }
	virtual void toggleActive() { active ? active = false : active = true; }

	/** returns the sum of the pixels color values
	 */
	virtual aggregateT getSum() const {
		aggregateT sum = 0;
		for (int i = 0; i < dim; i++) {
			sum += colors[i];
		}
		return sum;
	}

	virtual ChannelT getMax() {
		ChannelT max = -1;
		for (int i = 0; i < dim; i++) {
			if (colors[i] > max) {
				max = colors[i];
			}
		}
		return max;
	}

	virtual ChannelT getMin() {
		ChannelT min = -1;
		for (int i = 0; i < dim; i++) {
			if (colors[i] > min) {
				min = colors[i];
			}
		}
		return min;
	}

	virtual int getMaxPos() {
		ChannelT max = -1;
		int pos = -1;
		for (int i = 0; i < dim; i++) {
			if (colors[i] > max) {
				max = colors[i];
				pos = i;
			}
		}
		return pos;
	}

	virtual int getMinPos() {
		ChannelT min = -1;
		int pos = -1;
		for (int i = 0; i < dim; i++) {
			if (colors[i] > min) {
				min = colors[i];
				pos = i;
			}
		}
		return pos;
	}


	/** equality test for two pixels. Note that two pixels do not need to
	    have the same dimension in order to be considered as equal, as long
	    as in the remaining dimensions the values are set to zero.
	    Note also that the right-hand pixel is statically cast to the
	    lefthand-side, which might (in the case of different data types) also 
	    lead to inprecisions (-> overwrite this method at will).
	    */
	bool operator==(Pixel &p) {
		// FIXME: is it possible to ask for a pixel of arbitrary type ChannelT
		// as an argument of this function?
		int dim2 = p.getDim();
		if (dim >= dim2) {
			for (int i = 0; i < dim2; i++) if (colors[i] != static_cast<ChannelT>(p[i])) return false;
			for (int i = dim2; i < dim; i++) if (colors[i] != 0) return false;
		} else {
			for (int i = 0; i < dim; i++) if (colors[i] != static_cast<ChannelT>(p[i])) return false;
			for (int i = dim; i < dim2; i++) if (p[i] != 0) return false;
		}
		return true;
	}

	/** inequality test for two pixels.
	*/
	bool operator!=(Pixel &p) {
		return !operator==(p);
	}

	// this pixel is strictly lower than the other iff a lower color channel is strictly lower (-> lexicographical order; e.g. (100,101,102) < (102,101,100).
	bool operator<(const Pixel &p) const {
		assert(dim == p.getDim());
		const ChannelT *otherCol = p.getColors();
		for (int i = 0; i < dim; i++) {
			if (colors[i] < otherCol[i]) return true;
			if (colors[i] > otherCol[i]) return false;
		}
		return false;
	}

protected:
	ChannelT colors[dim];
	bool active;
};


template<typename T>
class Pixel3 : public Pixel<T, 3> {
public:
	Pixel3() : Pixel<T, 3>() {
	}
	Pixel3(T a, T b, T c) : Pixel<T, 3>() {
		this->colors[0] = a; this->colors[1] = b; this->colors[2] = c;
	}
};


#endif //PIXEL_H
